#pragma once
#include <list>
#include <string>
#include <vector>
#include <memory>
#include <mosquitto.h>
#include <iostream>

class IMQTTObserver
{
public:
	virtual ~IMQTTObserver ( )
	{
	}
	virtual void OnConnect ( int rc )=0;
	virtual void OnMessage ( const struct mosquitto_message *message )=0;
	virtual void OnSubscribe ( int mid, int qos_count, const int *granted_qos )=0;
};

typedef std::shared_ptr <IMQTTObserver> PMQTTObserver;

class TMQTTClientBase
{
public:
	virtual ~TMQTTClientBase ( )
	{
	}
	virtual void Connect ( ) = 0;
	virtual int Publish ( int *mid, const std::string &topic, const std::string &payload = "", int qos = 0, bool retain = false ) = 0;
	virtual int Subscribe ( int *mid, const std::string &sub, int qos = 0 ) = 0;
	virtual std::string Id ( ) const = 0;
	void Observe ( PMQTTObserver observer )
	{
		Observers.push_back ( observer );
	}
	void Unobserve ( PMQTTObserver observer )
	{
		Observers.remove ( observer );
	}

protected:
	const std::list <PMQTTObserver>& GetObservers ( ) const
	{
		return Observers;
	}

private:
	std::list <PMQTTObserver> Observers;
};

typedef std::shared_ptr <TMQTTClientBase> PMQTTClientBase;

class TMQTTClient : public TMQTTClientBase
{
public:

	struct TConfig
	{
		int Port;
		std::string Host;
		int Keepalive;
		std::string Id;

		TConfig ( ) : Port ( 1883 ), Host ( "localhost" ), Keepalive ( 60 ), Id ( "" )
		{
		}
	};

	TMQTTClient ( const TConfig &config );
	~TMQTTClient ( );

	virtual void on_connect ( int rc );
	virtual void on_message ( const struct mosquitto_message *message );
	virtual void on_subscribe ( int mid, int qos_count, const int *granted_qos );

	inline void Authenticate ( const std::string &user, const std::string &password )
	{
		if ( !user.empty ( ) )
		{
			auto retVal = mosquitto_username_pw_set ( mqtt, user.c_str ( ), password.c_str ( ) );
			if ( retVal != MOSQ_ERR_SUCCESS )
			{
				std::cerr << "username_pw_set: " << retVal << std::endl;
				exit ( retVal );
			}
		}
	}

	//	Connect
	void Connect ( ) override;

	//	Disconnect
	virtual void Disconnect ( );

	//	Publish
	int Publish ( int *mid, const std::string &topic, const std::string &payload = "", int qos = 0, bool retain = false );

	// Subscribe
	int Subscribe ( int *mid, const std::string &sub, int qos = 0 );

	//	Unsubscribe
	void Unsubscribe ( int* mid, const std::string& topic );

	std::string Id ( ) const
	{
		return MQTTConfig.Id;
	}

protected:
	const TConfig MQTTConfig;

private:
	std::list <PMQTTObserver> Observers;
	mosquitto* mqtt;
	bool IsConnected;

	//	on_connect
	static void on_vconnect ( mosquitto* mqtt, void* context, int rc );
	//	on_disconnect
	static void on_vdisconnect ( mosquitto* mqtt, void* context, int rc );
	//	on_message
	static void on_vmessage ( mosquitto* mqtt, void* context, const mosquitto_message* msg );
	//	on_subscribe
	static void on_vsubscribe ( mosquitto* mqtt, void* context, int mid, int count, const int* qosList );

};

typedef std::shared_ptr <TMQTTClient> PMQTTClient;

//	vagner - не компилируется под weezy-armel
//using Base = TMQTTClient;

class TMQTTPrefixedClient : public TMQTTClient
{
	//	vagner - не компилируется под weezy-armel
	using Base = TMQTTClient;
	public:
	TMQTTPrefixedClient ( std::string prefix, const TConfig &config )
	: TMQTTClient ( config )
	                , Prefix ( std::move ( prefix ) )
	{
	}

	//	vagner - не компилируется под weezy-armel
	int Publish ( int *mid, const std::string &topic, const std::string &payload = "", int qos = 0, bool retain = false )
	                override;
	int Subscribe ( int *mid, const std::string &sub, int qos = 0 ) override;
	void on_message ( const struct mosquitto_message *message ) override;
	/*
	 int Publish ( int *mid, const std::string& topic, const std::string& payload = "", int qos = 0, bool retain = false );
	 int Subscribe ( int *mid, const std::string& sub, int qos = 0 );
	 void on_message ( const struct mosquitto_message *message );
	 */

private:
	std::string Prefix;
};

class TMQTTWrapper : public TMQTTClient, public IMQTTObserver,
                public std::enable_shared_from_this <TMQTTWrapper>
{
public:
	TMQTTWrapper ( const TConfig &config )
	: TMQTTClient ( config )
	{
	}
	void Init ( )
	{
		Observe ( shared_from_this ( ) );
	}
};

class TMQTTPrefixedWrapper : public TMQTTPrefixedClient, public IMQTTObserver,
                public std::enable_shared_from_this <TMQTTPrefixedWrapper>
{
	//	vagner - не компилируется под weezy-armel
	using Base = TMQTTPrefixedClient;
	public:
	TMQTTPrefixedWrapper ( std::string prefix, const TConfig &config )
	: Base ( prefix, config )
	{
	}

	void Init ( )
	{
		Observe ( shared_from_this ( ) );
	}
};
